#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce :
@File      : model.py
@Time      : 31.08.21 19:12
@Author    : flowmeadow
"""

from typing import List, Union

import numpy as np
from pyglet.gl import *
from glpg_flowmeadow.rendering.gpu.shader import Shader
from glpg_flowmeadow.rendering.gpu.vao import VAO
from glpg_flowmeadow.transformations.methods import compute_normals, flip_inside_out, rotate_vec


class Model:

    default_shader = "blinn_phong"
    default_color = (0.0, 0.0, 1.0)

    def __init__(
        self,
        vertices: np.ndarray,
        indices: np.ndarray,
        color: Union[np.ndarray, List[float]] = default_color,
        rotation=(1.0, 0.0, 0.0, 0.0),
        scale=(1.0, 1.0, 1.0),
        translation=(0.0, 0.0, 0.0),
        shader_name: Union[str, List[str]] = default_shader,
        inside_out=False,
        num_lights=1,
    ) -> None:
        """
        :param vertices: object vertex coordinates (m, 3) [float]
        :param indices: object triangle vertex indices (n, 3) [int]
        :param color: color of the object (3,) or each individual vertex (n, 3)
        :param rotation: initial rotation parameter; rotation axis [0, 1, 2] and rotation angle [3] in degrees
        :param scale: initial scale parameter; one for each axis (x, y, z)
        :param translation: initial translation vector
        :param shader_name: name of the shader or list of shader names for change during runtime
        :param inside_out: if True, flip face normals
        :param num_lights: number of light sources in scene
        """

        # color
        colors = self.format_color_array(color, vertices.shape[0])
        if inside_out:
            indices = flip_inside_out(indices)

        # set initial vertex positions
        vertices = rotate_vec(vertices, rotation[:3], rotation[3])  # rotate
        vertices *= np.array(scale)  # scale
        vertices += np.array(translation)  # translate

        # compute normals from vertices
        normals = compute_normals(indices, vertices)
        indices, vertices = np.array(indices, dtype=np.uint32), np.array(vertices, dtype=np.float32)
        self._indices = indices
        self._vertices = vertices
        self._normals = normals
        self._colors = colors

        # define VAO
        self.vao = VAO(indices, object_id=GL_TRIANGLES)
        self.vao.set_vbo("position", vertices)
        self.vao.set_vbo("color", colors[:, :3])
        self.vao.set_vbo("normal", normals)

        # initialize model matrix and model operations list
        self.model_matrix = np.identity(4, dtype=np.float32)
        self.operations = []

        # initialize shader
        self.shaders = dict()
        if isinstance(shader_name, str):
            shader_name = [shader_name]
        for s in shader_name:
            self.shaders[s] = Shader(s, num_lights=num_lights)
        self.first_shader = shader_name[0]

    def rotate(self, angle: float, x: float, y: float, z: float):
        """
        rotate model around an axis defined by x, y and z.
        :param angle: rotation angle
        :param x: x-factor
        :param y: y-factor
        :param z: z-factor
        """

        rot_vec = [angle, x, y, z]

        # normalize axis
        axis = np.array([x, y, z])
        l, m, n = axis / np.linalg.norm(axis)

        # compute angle, sine and cosine values
        angle = np.pi * angle / 180.0
        s, c = np.sin(angle), np.cos(angle)

        # build transformation matrix (4, 4)
        rot_transform = np.array(
            [
                [l * l * (1 - c) + c, m * l * (1 - c) - n * s, n * l * (1 - c) + m * s, 0.0],
                [l * m * (1 - c) + n * s, m * m * (1 - c) + c, n * m * (1 - c) - l * s, 0.0],
                [l * n * (1 - c) - m * s, m * n * (1 - c) + l * s, n * n * (1 - c) + c, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ]
        )
        # update model matrix and operations list
        self.model_matrix = self.model_matrix @ rot_transform
        self.operations.append([glRotatef, rot_vec])

    def scale(self, x: float, y: float, z: float):
        """
        scale model in x, y and z direction
        :param x: x-factor
        :param y: y-factor
        :param z: z-factor
        """
        scale_vec = [x, y, z]

        # build transformation matrix (4, 4)
        scale_transform = np.identity(4, dtype=np.float)
        np.fill_diagonal(scale_transform[:3, :3], scale_vec)

        # update model matrix and operations list
        self.model_matrix = self.model_matrix @ scale_transform
        self.operations.append([glScalef, scale_vec])

    def translate(self, x: float, y: float, z: float):
        """
        translate model about vector (x, y, z)
        :param x: x-factor
        :param y: y-factor
        :param z: z-factor
        """
        trans_vec = [x, y, z]

        # build transformation matrix (4, 4)
        trans_transform = np.identity(4, dtype=np.float)
        trans_transform[:-1, 3] = trans_vec

        # update model matrix and operations list
        self.model_matrix = self.model_matrix @ trans_transform
        self.operations.append([glTranslatef, trans_vec])

    def reset(self):
        """
        Reset model
        """
        self.model_matrix = np.identity(4, dtype=np.float32)
        self.operations = []

    def draw(self, shader_name: str = None, num_indices: int = None, offset=0):
        """
        Update shader, perform all transformations and draw model
        :param shader_name: select a shader
        :param num_indices: number of indices to draw; by default all
        :param offset: start point of the indices to draw; by default first (0)
        """

        # update shader
        if shader_name is None:
            shader_name = self.first_shader
        self.update_shader(shader_name)
        self.shaders[shader_name].use()

        glPushMatrix()
        # perform transformations
        for operation, args in self.operations:
            operation(*args)
        # draw model
        self.vao.draw(num_indices, offset)
        glPopMatrix()
        glFlush()

        # reset
        self.shaders[shader_name].un_use()
        self.reset()

    def update_shader(self, shader_name: str):
        """
        Updates shader attributes
        :param shader_name: shader to update
        """
        self.shaders[shader_name].update_model_matrix(self.model_matrix)
        self.shaders[shader_name].update_time()

    @staticmethod
    def format_color_array(color: np.ndarray, num_vertices: int) -> np.ndarray:
        """
        preprocess a given color array and bring it to size (m, 3)
        :param color: the original color array
        :param num_vertices: number of vertices (m)
        :return: color array (m, 3)
        """
        # type checking
        if isinstance(color, (list, tuple)):
            color = np.array(color)
        elif not isinstance(color, np.ndarray):
            raise ValueError("Color has to be a tuple, list or np.ndarray")
        # dimension checking
        if color.ndim == 1 and color.shape[0] == 3:
            color = np.full((num_vertices, 3), color)  # bring array to size (m, 3)
        elif color.ndim == 2 and color.shape[1] == 3:
            pass
        else:
            raise ValueError("Color array has to be 1 or 2 dimensional with each color having 3 entries")
        return color

    def update_camera(self, *args, **kwargs):
        """
        update camera parameters for every shader
        :param args: forwarded arguments
        :param kwargs: forwarded keyword arguments
        """
        for key in self.shaders.keys():
            self.shaders[key].update_camera(*args, **kwargs)

    def update_object(self, vertices: np.ndarray = None, colors: np.ndarray = None):
        """
        Update vertices and/or colors of the model
        :param vertices: vertices array (m, 3)
        :param colors: color array (m, 3) pr (3,)
        """
        if vertices is not None:
            self._vertices = np.array(vertices, dtype=np.float32)
            self._normals = compute_normals(self._indices, self._vertices)
            self.vao.set_vbo("normal", self._normals)
            self.vao.set_vbo("position", self._vertices)

        if colors is not None:
            self._colors = self.format_color_array(colors, self._vertices.shape[0])
            self.vao.set_vbo("color", colors[:, :3])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : Simple point cloud object
@File      : point_cloud.py
@Project   : pygletPlayground
@Time      : 30.03.22 23:12
@Author    : flowmeadow
"""
from typing import Optional

import numpy as np
from pyglet.gl import *


class PointCloud:
    """
    Simple point cloud object
    """

    def __init__(self, points: np.ndarray, colors: Optional[np.ndarray] = None, point_size: int = 10):
        """
        Object that draws several 3D points as dots into the scene
        :param points: point coordinates (m, 3)
        :param colors: point colors (m, 3)
        :param point_size: point size in pixels
        """
        # define vertices and colors
        self.points = points
        self.colors = np.full((1.0, 1.0, 1.0), self.points.shape[0]) if colors is None else colors
        self.point_size = point_size

    def draw(self):
        """
        draw the points into the scene
        :return:
        """
        glEnable(GL_POINT_SMOOTH)
        glPointSize(self.point_size)
        glBegin(GL_POINTS)
        for p, c in zip(self.points, self.colors):
            glColor3d(c[0], c[1], c[2])
            glVertex3d(p[0], p[1], p[2])
        glEnd()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : Handles loading, linking and use of GLSL shaders
@File      : shader.py
@Time      : 31.08.21 23:49
@Author    : flowmeadow
"""
import ctypes as ct
import os
import time

import numpy as np
from pyglet.gl import *
import importlib



class Shader:
    """
    Handles loading, linking and use of GLSL shaders
    """

    def __init__(self, shader_name: str, model_base: str = "base", num_lights: int = 1):
        """
        Create program and compile shader
        :param shader_name: filename of the shader without extension. A '*.vert' and '*.frag' file must be available.
        :param model_base: parent directory of the shader. Other models might require different shaders (e.g. material)
        :param num_lights: number of light sources
        """
        self.start_time = time.time()  # define time constant

        self.model_base = model_base
        self.program = glCreateProgram()

        # file_path = f"{CURRENT_PATH}/../../shader/{model_base}/{shader_name}/{shader_name}"
        # # load shader file text
        # with open(f"{file_path}.vert", "r") as f:
        #     vs_src = f.readlines()
        # with open(f"{file_path}.frag", "r") as f:
        #     fs_src = f.readlines()

        # load shader text
        shader_txt = importlib.import_module(f"glpg_flowmeadow.shader.{model_base}.{shader_name}.{shader_name}")
        vs_src = shader_txt.vert_txt
        fs_src = shader_txt.frag_txt

        # compile shader
        vs = self.load_shader(vs_src, GL_VERTEX_SHADER, shader_name)
        if not vs:
            raise ValueError("Vertex shader could not be loaded")
        fs = self.load_shader(fs_src, GL_FRAGMENT_SHADER, shader_name)
        if not fs:
            raise ValueError("Fragment shader could not be loaded")

        # compile program
        self.program = glCreateProgram()
        glAttachShader(self.program, vs)
        glAttachShader(self.program, fs)

        # link program
        glLinkProgram(self.program)

        # not needed anymore
        glDeleteShader(vs)
        glDeleteShader(fs)

        # set uniforms
        self.model_matrix_loc = glGetUniformLocation(self.program, "modelMatrix".encode("ascii"))
        self.camera_pos_loc = glGetUniformLocation(self.program, "cameraPos".encode("ascii"))
        self.camera_view_loc = glGetUniformLocation(self.program, "cameraView".encode("ascii"))

        self.time_loc = glGetUniformLocation(self.program, "iTime".encode("ascii"))

        num_lights_loc = glGetUniformLocation(self.program, "iLights".encode("ascii"))

        # start configuration
        self.use()
        glUniform1ui(num_lights_loc, ct.c_uint(num_lights))  # set number of light sources
        self.un_use()

    def use(self) -> None:
        """
        Select the current program
        """
        glUseProgram(self.program)

    @staticmethod
    def un_use() -> None:
        """
        Unselect the current program
        """
        glUseProgram(0)

    @staticmethod
    def load_shader(src: str, shader_type: int, file_name: str = None) -> int:
        """
        Compile a shader
        :param src: text of the shader file
        :param shader_type: type of the shader (GL_VERTEX_SHADER or GL_FRAGMENT_SHADER)
        :param file_name: file_name of the shader file. Used for debugging only
        :return: shader id
        """
        extensions = {GL_FRAGMENT_SHADER: "frag", GL_VERTEX_SHADER: "vert"}

        # convert source text
        c_text = (ct.c_char_p * len(src))(*[line.encode("utf-8") for line in src])

        # create shader
        shader = glCreateShader(shader_type)
        glShaderSource(shader, len(src), ct.cast(ct.pointer(c_text), ct.POINTER(ct.POINTER(ct.c_char))), None)

        # compile shader
        glCompileShader(shader)

        # check shader status
        status = ct.c_int(0)
        glGetShaderiv(shader, GL_COMPILE_STATUS, ct.byref(status))

        if not status:
            # retrieve the log text
            glGetShaderiv(shader, GL_INFO_LOG_LENGTH, ct.byref(status))
            log = ct.create_string_buffer(status.value)
            glGetShaderInfoLog(shader, status, None, log)
            log_text = log.value.decode("utf-8")

            glDeleteShader(shader)

            file_name = file_name if file_name else "unknown"
            raise ImportError(f"{file_name}.{extensions[shader_type]}:\n\n{log_text}")
        return shader

    def update_model_matrix(self, matrix: np.array):
        """
        Updates the uniform modelMatrix attribute
        :param matrix: transformation matrix [4, 4]
        """
        self.use()
        mat = matrix.flatten("F")
        glUniformMatrix4fv(self.model_matrix_loc, 1, False, (ct.c_float * 16)(*mat))
        self.un_use()

    def update_camera(self, camera_pos: np.ndarray, camera_view: np.ndarray):
        """
        update camera attributes
        :param camera_pos: camera position array (3,)
        :param camera_view: camera view direction array (3,)
        """
        self.use()
        glUniform3f(self.camera_pos_loc, *np.array(camera_pos, dtype=np.float32).flatten("F"))
        glUniform3f(self.camera_view_loc, *np.array(camera_view, dtype=np.float32).flatten("F"))
        self.un_use()

    def update_time(self):
        """
        update runtime in milliseconds
        """
        self.use()
        t = int(1000 * (time.time() - self.start_time))
        glUniform1ui(self.time_loc, t)
        self.un_use()



vert_txt = """
#version 130

uniform mat4 modelMatrix;

varying vec3 fNormal;
varying vec3 fPosition;
varying vec3 fColor;

void main ()
{
    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
    fPosition = vec3(modelMatrix * gl_Vertex);
    fColor = vec3(gl_Color);
    fNormal = normalize(mat3(modelMatrix) * gl_Normal);
}
"""

frag_txt = """
#version 130

out vec4 FragColor;

uniform vec3 cameraPos;
uniform uint iLights;

varying vec3 fNormal;
varying vec3 fPosition;
varying vec3 fColor;

#define STEPS 2
#define CONTOUR_THICKNESS 8.8  // [0, 10]

float attenuation(int light_index, vec3 light_dir){
    float d = abs(length(light_dir));
    float attenuation = gl_LightSource[light_index].constantAttenuation;
    attenuation += gl_LightSource[light_index].linearAttenuation * d;
    attenuation += gl_LightSource[light_index].quadraticAttenuation * pow(d, 2);
    attenuation = 1. / attenuation;
    return attenuation;
}

void main()
{

    vec3 ambient_color  = 0.3  * fColor;
    vec3 final_color = vec3(0., 0., 0.);

    vec3 v = normalize(cameraPos - fPosition);
    vec3 n = normalize(fNormal);

    for (int l_idx = 0; l_idx < int(iLights); l_idx++){
        vec3 light_dir = vec3(gl_LightSource[l_idx].position) - fPosition;
        vec3 l = normalize(light_dir);
        vec3 h = (v + l) / length(v + l);

        float spot = (1. - dot(-l, gl_LightSource[l_idx].spotDirection)) * 90.;
        spot = max(sign(gl_LightSource[l_idx].spotCutoff - spot), 0.);

        float diffuse_factor = round(max(0.0, dot(n, l)) * STEPS) / STEPS;
        vec3 diffuse_color = 0.3 * diffuse_factor * vec3(gl_LightSource[l_idx].diffuse);
        float specular_factor = round(pow(max(0.0, dot(n, h)), 64) * STEPS) / STEPS ;
        vec3 specular_color = 0.5 * specular_factor * vec3(gl_LightSource[l_idx].specular);

        final_color += (spot * (diffuse_color + specular_color) + ambient_color / iLights) * attenuation(l_idx, light_dir);
    }

    // draw black contour; value is either 0. or 1.
    float contour_factor = min(round((10. - CONTOUR_THICKNESS) * dot(v, n)), 1.);
    FragColor = vec4(final_color * contour_factor, 1.);
}
"""


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : Manages several OpenGL light sources
@File      : lights.py
@Time      : 15.09.21 16:47
@Author    : flowmeadow
"""

from pyglet.gl import *

from glpg_flowmeadow.rendering.lighting.light import Light


class Lights:
    """
    Manages several OpenGL light sources
    """
    _light_ids = {
        0: GL_LIGHT0,
        1: GL_LIGHT1,
        2: GL_LIGHT2,
        3: GL_LIGHT3,
        4: GL_LIGHT4,
        5: GL_LIGHT5,
        6: GL_LIGHT6,
        7: GL_LIGHT7,
    }

    def __init__(self):
        """ Initialize """
        self.num_lights = 0
        self.lights = []

    def add(self, **kwargs):
        """
        Adds a new light source
        :param kwargs: forwarded keyword arguments
        """
        # create a new light object
        light_id = self._light_ids[self.num_lights]
        self.num_lights += 1
        self.lights.append(Light(light_id, **kwargs))

    def draw(self):
        """
        Draws a representation of all light sources into the scene
        """
        for light in self.lights:
            light.draw()

    def __getitem__(self, idx: int) -> Light:
        """
        Returns a light object by its index
        :param idx: Index of the light source
        :return: light object
        """
        return self.lights[idx]






#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : several methods to draw simple references into a scene
@File      : rendering.py
@Time      : 09.09.21 00:23
@Author    : flowmeadow
"""
from typing import Tuple

import numpy as np
import pyglet
from pyglet.gl import *

# def draw_text_3D(x_txt: int, y_txt: int, z_txt: int, text: str, color=(1.0, 1.0, 1.0)) -> None:
#     """
#     Draws a given string into the scene.
#     :param x_txt: x-position
#     :param y_txt: y_txt-position
#     :param z_txt: z_txt-position
#     :param text: test to draw
#     :param color: color of the text
#     :return: None
#     """
#     font = GLUT_BITMAP_8_BY_13
#     glColor3f(color[0], color[1], color[2])
#     glRasterPos3f(x_txt, y_txt, z_txt)
#     for ch in text:
#         glutBitmapCharacter(font, ord(ch))


def draw_text_2D(x_txt: int, y_txt: int, text: str) -> None:
    """
    Draw a text onto the screen
    :param x_txt: x-position
    :param y_txt: y-position
    :param text: text string
    :return:
    """
    label = pyglet.text.Label(
        text, font_name="Consolas", font_size=12, x=x_txt, y=y_txt, anchor_x="left", anchor_y="top"
    )
    label.draw()


def draw_coordinates(R=np.eye(3), t=np.zeros(3), scale=1.0):
    """
    Draws a reference coordinate system in the scene
    :param R: rotation matrix array (3, 3)
    :param t: center array of the coordinate system (3,)
    :param scale: scale factor
    """
    center = np.array([0.0, 0.0, 0.0]) + t
    vertices = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    )
    colors = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    )
    glBegin(GL_LINES)
    for v, c in zip(vertices, colors):
        v = (R @ v) * scale + t
        glColor3f(*c)
        glVertex3f(*center)
        glVertex3f(*v)
    glEnd()
    glFlush()


def draw_light(light_pos: np.ndarray):
    """
    Draw a reference for a point light source
    :param light_pos: center of the light source (3,)
    :return: None
    """
    glColor3f(1.0, 1.0, 0.0)
    glBegin(GL_LINES)
    glVertex3f(light_pos[0] - 0.1, light_pos[1], light_pos[2])
    glVertex3f(light_pos[0] + 0.1, light_pos[1], light_pos[2])
    glVertex3f(light_pos[0], light_pos[1] - 0.1, light_pos[2])
    glVertex3f(light_pos[0], light_pos[1] + 0.1, light_pos[2])
    glVertex3f(light_pos[0], light_pos[1], light_pos[2] - 0.1)
    glVertex3f(light_pos[0], light_pos[1], light_pos[2] + 0.1)
    glVertex3f(0.0, 0.0, 1.0)
    glEnd()
    glFlush()


def draw_cam(
    pos: np.ndarray,
    rot_mat: np.ndarray,
    radius: float,
    side_num=20,
    edge_only=True,
    color: Tuple[float, float, float] = None,
):
    """
    Draw a reference of a camera position based on a cone
    :param pos: position of the camera (3,)
    :param rot_mat: rotation matrix of the camera (3, 3)
    :param radius: cone radius
    :param side_num: number of sides to represent the cone
    :param edge_only: if True, only draw the polygon edges
    :param color: color of the cone
    """
    # prepare data
    color = (1.0, 1.0, 1.0) if color is None else color
    glColor3f(*color)
    if edge_only:
        glBegin(GL_LINE_LOOP)
    else:
        glBegin(GL_POLYGON)
    vertices = []

    # build transformation matrix
    T = np.eye(4)
    T[:3, :3] = rot_mat
    T[:3, -1] = pos

    # build circle vertices and connect them
    for idx in range(0, side_num):
        angle = idx * 2.0 * np.pi / side_num
        vertex = np.array([np.cos(angle) * radius, np.sin(angle) * radius, 0.5])
        vertex = (T @ np.append(vertex, 1.0))[:3]
        vertices.append(vertex)
        glVertex3f(*vertex)
    glEnd()

    # connect circle vertices with camera center
    glBegin(GL_LINES)
    for vertex in vertices:
        glVertex3f(*pos)
        glVertex3f(*vertex)
    glEnd()

    # draw a small coordinate system for reference
    draw_coordinates(rot_mat, pos, scale=0.2)
    glFlush()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : demo script
@File      : demo.py
@Project   : pygletPlayground
@Time      : 01.10.21 23:50
@Author    : flowmeadow
"""
import numpy as np

from glpg_flowmeadow.camera.fly_motion import FlyMotion
from glpg_flowmeadow.display.gl_screen import GLScreen
from glpg_flowmeadow.rendering.lighting.lights import Lights
from glpg_flowmeadow.rendering.methods import draw_coordinates, draw_text_2D
from glpg_flowmeadow.rendering.models.model import Model
from glpg_flowmeadow.rendering.models.model_generation.geometry import sphere, cube, cylinder


class MyScreen(GLScreen):
    """
    Demo application class containing a movable camera, three light sources and a
    sphere model that is transformed and shaded dynamically
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.frame_count = 0

        # define shader list to use
        self.shaders = ["flat", "gouraud", "blinn_phong", "toon"]
        self.current_shader = self.shaders[0]

        # define camera
        self.cam = FlyMotion(self, camera_pos=(1.0, 1.0, 1.0), camera_view=(-1.0, -1.0, -1.0))

        # define light sources
        self.lights = Lights()
        self.lights.add(  # circling around
            position=(1.0, 1.0, 0.0),
            ambient=(1.0, 1.0, 0.5),
            diffuse=(1.0, 1.0, 0.5),
            specular=(1.0, 1.0, 0.5),
        )
        self.lights.add(  # at camera position
            position=(0.0, 0.0, 1.0),
            ambient=(0.0, 0.5, 1.0),
            diffuse=(0.0, 0.5, 1.0),
            specular=(0.0, 0.5, 1.0),
        )
        self.lights.add(  # constant
            position=(0.0, 0.0, 1.0),
            ambient=(1.0, 0.0, 0.5),
            diffuse=(1.0, 0.0, 0.5),
            specular=(1.0, 0.0, 0.5),
        )

        # define sphere model
        vertices, indices = sphere(refinement_steps=4)
        self.sphere = Model(
            vertices=vertices,
            indices=indices,
            shader_name=self.shaders,
            color=(0.0, 0.4, 1.0),
            rotation=(1.0, 0.0, 0.0, 90.0),
            translation=(0.0, 0.0, 0.1),
            num_lights=self.lights.num_lights,
        )
        vertices, indices = cube(refinement_steps=3)
        self.cube = Model(
            vertices=vertices,
            indices=indices,
            shader_name=self.shaders,
            color=(0.6, 0.4, 0.1),
            scale=0.1,
            rotation=(0.0, 0.0, 1.0, 45.0),
            num_lights=self.lights.num_lights,
        )

        vertices, indices = cylinder()
        self.cylinder = Model(
            vertices=vertices,
            indices=indices,
            shader_name=self.shaders,
            color=(0.2, 0.8, 0.1),
            translation=(0.0, 0.5, 0.0),
            num_lights=self.lights.num_lights,
        )

    def handle_events(self) -> None:
        """
        handle pygame events and do other stuff before drawing
        :return: None
        """
        self.sphere.update_camera(self.cam.camera_pos, self.cam.camera_view)  # update camera based on user inputs
        pass

    def draw_world(self) -> None:
        """
        draw objects in the world
        :return: None
        """
        draw_coordinates()

        # update shader
        idx = (self.frame_count // 500) % len(self.shaders)
        self.current_shader = self.shaders[idx]

        # move light sources
        new_pos = (np.sin(self.frame_count * 0.01), np.cos(self.frame_count * 0.01), 0.5)
        self.lights[0].update("position", new_pos)
        self.lights[1].update("position", self.cam.camera_pos)

        self.lights.draw()

        # transform sphere
        z = np.abs(np.sin(self.frame_count * 0.01)) / 2
        self.sphere.translate(0.0, 0.0, z)
        self.sphere.draw(shader_name=self.current_shader)

        # transform cube
        self.cube.draw(shader_name=self.current_shader)
        self.cube.translate(0.5, 0.0, 0.1)
        self.cube.rotate(self.frame_count * 0.2, 0.0, 0.0, 1.0)

        self.cylinder.draw(shader_name=self.current_shader)
        self.cylinder.scale(1.0, 1.0, 1.0 - z)

        # update frame counter
        self.frame_count += 1

    def draw_screen(self) -> None:
        """
        draw objects onto the screen
        :return: None
        """
        draw_text_2D(10, self.height - 10, f"FPS: {self.current_fps:.2f}")
        draw_text_2D(self.width - 300, self.height - 10, f"Current Shader: {self.current_shader.upper()}")
        draw_text_2D(self.width - 300, self.height - 35, f"Next in {(500 - self.frame_count) % 500} frames")
        if self.frame_count < 250:
            text = "Use ASDF to move and MOUSE to look around"
            draw_text_2D(self.width / 2 - 200, self.height / 2, f"{text}")


if __name__ == "__main__":
    demo = MyScreen(fullscreen=True)
    demo.run()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : TODO WIP
@File      : geometry.py
@Time      : 10.09.21 10:50
@Author    : flowmeadow
"""
from typing import Tuple

import numpy as np


def cube(size=1, refinement_steps=0):
    vertices = [
            # corners
            [-1.0, -1.0, -1.0],
            [-1.0, -1.0, +1.0],
            [-1.0, +1.0, -1.0],
            [-1.0, +1.0, +1.0],
            [+1.0, -1.0, -1.0],
            [+1.0, -1.0, +1.0],
            [+1.0, +1.0, -1.0],
            [+1.0, +1.0, +1.0],
            # face centers
            [+0.0, +0.0, -1.0],
            [+0.0, +0.0, +1.0],
            [+0.0, -1.0, +0.0],
            [+0.0, +1.0, +0.0],
            [-1.0, +0.0, +0.0],
            [+1.0, +0.0, +0.0],
    ]
    indices = [
            # face 1
            [8, 0, 2],
            [8, 2, 6],
            [8, 6, 4],
            [8, 4, 0],
            # face 2
            [9, 3, 1],
            [9, 7, 3],
            [9, 5, 7],
            [9, 1, 5],

            [10, 1, 0],
            [10, 5, 1],
            [10, 4, 5],
            [10, 0, 4],

            [11, 2, 3],
            [11, 6, 2],
            [11, 7, 6],
            [11, 3, 7],

            [12, 0, 1],
            [12, 2, 0],
            [12, 3, 2],
            [12, 1, 3],

            [13, 5, 4],
            [13, 7, 5],
            [13, 6, 7],
            [13, 4, 6],
    ]
    vertices_per_face = 3
    for step in range(refinement_steps):
        print(f"Refinement step: ({step + 1}|{refinement_steps})")
        new_vertices, new_indices = vertices, []
        vertices = np.array(vertices)
        for triangle in indices:
            edges = []
            for i in range(vertices_per_face):
                edges.append(vertices[triangle[(i + 1) % vertices_per_face]] - vertices[triangle[i]])
            edges = np.array(edges)
            largest_edge = np.argmax(np.linalg.norm(edges, axis=1))
            new_vertex = vertices[triangle[largest_edge]] + edges[largest_edge] * 0.5

            if new_vertex.tolist() not in new_vertices:
                vertex_idx = len(new_vertices)
                new_vertices.append(new_vertex.tolist())
            else:
                vertex_idx = new_vertices.index(new_vertex.tolist())

            triangle_1 = [triangle[0], triangle[1], triangle[2]]
            triangle_2 = [triangle[0], triangle[1], triangle[2]]

            triangle_1[largest_edge] = vertex_idx
            triangle_2[(largest_edge + 1) % vertices_per_face] = vertex_idx

            new_indices.append(triangle_1)
            new_indices.append(triangle_2)

        vertices, indices = new_vertices, new_indices

    return np.array(vertices) * size, np.array(indices)


def sphere(radius=0.1, refinement_steps=3):
    vertices = np.array([
        [0., 0., -radius],
        [0., radius, 0.],
        [radius, 0., 0.],
        [0., -radius, 0.],
        [-radius, 0., 0.],
        [0., 0., radius],
    ])
    indices = np.array([
        [0, 1, 2],
        [0, 2, 3],
        [0, 3, 4],
        [0, 4, 1],
        [5, 4, 3],
        [5, 3, 2],
        [5, 2, 1],
        [5, 1, 4],
    ])

    # refinement
    for _ in range(refinement_steps):
        new_indices, edges = [], []
        for triangle in indices:
            vert_index = len(vertices)
            new_vert = np.mean(vertices[triangle], axis=0)
            new_vert = (new_vert / np.linalg.norm(new_vert)) * radius
            vertices = np.append(vertices, [new_vert], axis=0)
            new_indices.append([vert_index, triangle[0], triangle[1]])
            new_indices.append([vert_index, triangle[1], triangle[2]])
            new_indices.append([vert_index, triangle[2], triangle[0]])
        indices = np.array(new_indices)

        # edge flipping
        for idx in range(len(indices)):
            triangle = indices[idx]
            c_triangle, c_idx = 0, 0
            for c_idx in range(len(indices)):
                c_triangle = indices[c_idx]
                if triangle[1] in c_triangle and triangle[2] in c_triangle and triangle[0] not in c_triangle:
                    break
            c_vert = c_triangle[np.where((c_triangle != triangle[1]) & (c_triangle != triangle[2]))[0]]

            edge_1 = [triangle[1], triangle[2]]
            edge_2 = [triangle[0], c_vert[0]]

            dist_1 = np.linalg.norm(vertices[edge_1[0]] - vertices[edge_1[1]])
            dist_2 = np.linalg.norm(vertices[edge_2[0]] - vertices[edge_2[1]])

            if dist_1 > dist_2:
                # perform edge flip
                indices[idx] = [edge_1[0], edge_2[1], edge_2[0]]
                indices[c_idx] = [edge_1[1], edge_2[0], edge_2[1]]
    return vertices, indices


def cylinder(radius: float = 0.1, height: float = 0.2, angle_steps: int = 32, height_steps: int = 8):
    angles = np.linspace(0, 2 * np.pi, angle_steps, endpoint=False)  # get values of all angles

    vertices, indices = [], []
    # create the vertices
    for h in np.linspace(0, height, height_steps):
        vertices.append([0, 0, h])
        for a in angles:
            vertices.append([np.sin(a) * radius, np.cos(a) * radius, h])

    # create the indices
    for i in range(1, angle_steps + 1):
        # top and bottom triangles
        indices.append([0, i, i % angle_steps + 1])
        offset = (angle_steps + 1) * (height_steps - 1)
        indices.append([offset, offset + i, offset + i % angle_steps + 1])
        # side triangles
        for h in range(height_steps - 1):
            offset_1 = (angle_steps + 1) * h
            offset_2 = (angle_steps + 1) * (h + 1)
            indices.append([offset_1 + i, offset_2 + i % angle_steps + 1, offset_1 + i % angle_steps + 1])
            indices.append([offset_1 + i, offset_2 + i, offset_2 + i % angle_steps + 1])
    return np.array(vertices), np.array(indices)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : several methods to handle vector transformations
@File      : transformations.py
@Time      : 09.09.21 23:26
@Author    : flowmeadow
"""
import ctypes
from typing import Tuple

import numpy as np
from pyglet.gl import *


def rot_mat(axis: np.ndarray, angle: float) -> np.ndarray:
    """
    build rotation matrix from axis and angle
    :param axis: axis to rotate around [3,]
    :param angle: angle of rotation in degrees
    :return: rotated vector
    """
    angle = np.pi * angle / 180.0
    l, m, n = axis / np.linalg.norm(axis)  # normalize axis vector
    s, c = np.sin(angle), np.cos(angle)
    # build rotation matrix
    rot_transform = np.array(
        [
            [l * l * (1 - c) + c, m * l * (1 - c) - n * s, n * l * (1 - c) + m * s],
            [l * m * (1 - c) + n * s, m * m * (1 - c) + c, n * m * (1 - c) - l * s],
            [l * n * (1 - c) - m * s, m * n * (1 - c) + l * s, n * n * (1 - c) + c],
        ]
    )
    return rot_transform


def rotate_vec(vec: np.ndarray, axis: np.ndarray, angle: float) -> np.ndarray:
    """
    Rotate a vector or an array of vectors along an axis given an angle
    :param vec: vector [3,] or array of vectors [n, 3] to rotate
    :param axis: axis to rotate around [3,]
    :param angle: angle of rotation in degrees
    :return: rotated vector
    """
    rot_transform = rot_mat(axis, angle)
    if vec.ndim == 1:
        result = rot_transform @ vec
    else:
        result = (rot_transform @ vec.T).T
    return result


def rotation_matrix_from_vectors(vec1: np.ndarray, vec2: np.ndarray):
    """
    Find the rotation matrix that aligns vec1 to vec2
    :param vec1: A 3d "source" vector
    :param vec2: A 3d "destination" vector
    :return mat: A transform matrix (3x3) which when applied to vec1, aligns it with vec2.
    """
    a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (vec2 / np.linalg.norm(vec2)).reshape(3)
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    k_mat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    rotation_matrix = np.eye(3) + k_mat + k_mat.dot(k_mat) * ((1 - c) / (s ** 2))
    return rotation_matrix


def compute_normals(indices: np.ndarray, vertices: np.ndarray) -> np.ndarray:
    """
    Given a vertex and index array, the normals for each vertex are computed.
    First the face normals will be calculated. For each vertex the normal results from
    the face normals of its surrounding faces.
    :param indices: indices array [m, 3]
    :param vertices: vertex array [n, 3]
    :return: normal array [n, 3]
    """
    # compute triangle normals by building the cross product of two edges
    triangle_vertices = vertices[indices]
    edge_1 = triangle_vertices[:, 0, :] - triangle_vertices[:, 1, :]
    edge_2 = triangle_vertices[:, 0, :] - triangle_vertices[:, 2, :]
    cross_products = np.cross(edge_1, edge_2)
    # normalize the normal vectors
    triangle_normals = cross_products / np.linalg.norm(cross_products, axis=1)[:, np.newaxis]

    # compute vertex normals
    # TODO: There has to be a more efficient way ...
    vertex_normals = np.zeros(vertices.shape)
    # add up all the surrounding face normals for each vertex
    for idx in range(len(indices)):
        n = triangle_normals[idx]
        f = indices[idx, :]
        vertex_normals[f] += n
    # normalize the normal vectors
    vertex_normals = vertex_normals / np.linalg.norm(vertex_normals, axis=1)[:, np.newaxis]
    return vertex_normals


def flip_inside_out(indices: np.array) -> np.array:
    """
    To shade the inside of the model, the logical order of face vertices has
    to be changed from counterclockwise to clockwise.
    :param indices: indices array [m, 3]
    :return: indices array [m, 3]
    """
    new_indices = indices.copy()
    new_indices[:, 1], new_indices[:, 2] = indices[:, 2], indices[:, 1]
    return new_indices


def get_P(matrix=GL_PROJECTION_MATRIX) -> np.ndarray:
    """
    :return: camera projection matrix as numpy array (4, 4)
    """
    # get projection matrix
    P = np.zeros(16)
    P = (ctypes.c_float * len(P))(*P)
    glGetFloatv(matrix, P)
    P = np.array(P).reshape((4, 4)).T
    return P


def get_M() -> np.ndarray:
    """
    :return: modelview matrix as numpy array (4, 4)
    """
    # get projection matrix
    M = np.zeros(16)
    M = (ctypes.c_float * len(M))(*M)
    glGetFloatv(GL_MODELVIEW_MATRIX, M)
    M = np.array(M).reshape((4, 4)).T
    return M


def get_K(P: np.ndarray, screen_size: Tuple[int, int]) -> np.ndarray:
    """

    :param P: camera projection matrix as numpy array (4, 4)
    :param screen_size: image screen dimension (width, height)
    :return: camera matrix K in OpenCV convention as numpy array (3, 3)
    """
    # compute K from projection matrix
    w, h = screen_size
    K = np.array(
        [
            [P[0, 0] * w / 2.0, 0.0, w * (1.0 - P[0, 2]) / 2.0],
            [0.0, -P[1, 1] * h / 2.0, h * (1.0 - P[1, 2]) / 2.0],
            [0.0, 0.0, 1.0],
        ]
    )
    return K


def construct_T(rot_mat: np.ndarray, trans_vec: np.ndarray) -> np.ndarray:
    """
    Combines rotation matrix and translation vector to transformation matrix
    :param rot_mat: array (3, 3)
    :param trans_vec: array (3,)
    :return: array (4, 4)
    """
    T = np.eye(4)
    T[:3, :3] = rot_mat
    T[:3, -1] = trans_vec
    return T



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : base camera class
@File      : camera.py
@Time      : 09.09.21 22:55
@Author    : flowmeadow
"""
from abc import ABCMeta, abstractmethod
from typing import Tuple, Union

import numpy as np
import pyglet.window

from glpg_flowmeadow.rendering.methods import draw_cam
from glpg_flowmeadow.transformations.methods import construct_T


class Camera:
    """
    base camera class containing the cameras pose information.
    'move' method has to be defined in the child classes
    """

    __metaclass__ = ABCMeta

    def __init__(
        self,
        camera_pos=(0.0, 0.0, 0.0),
        camera_view=(1.0, 0.0, 0.0),
        camera_up=(0.0, 0.0, 1.0),
        name="Camera",
    ):
        """
        :param camera_pos: position of the camera in world coordinates [3,]
        :param camera_view: view direction of the camera in world coordinates [3,]
        :param camera_up: vector pointing up from the camera world coordinates [3,]
        """
        self.name = name
        camera_view, camera_up = np.array(camera_view), np.array(camera_up)
        self._camera_pos = np.array(camera_pos)
        self._camera_view = camera_view / np.linalg.norm(camera_view)
        self._camera_up = camera_up / np.linalg.norm(camera_up)
        self._camera_up = self._update_camera_up()

    def _update_camera_up(self):
        nvec = np.cross(self.camera_view, np.array([0.0, 0.0, 1.0]))
        nvec = nvec / np.linalg.norm(nvec)
        uvec = np.cross(nvec, self.camera_view)
        return uvec

    def copy_from(self, cam: object):  # TODO: how to reference base class object from child for type hints
        """
        Copy the pose from another Camera object and
        :param cam: Camera object
        """
        self.camera_pos = cam.camera_pos
        self.camera_view = cam.camera_view
        self.camera_up = cam.camera_up

    def draw(self):
        """
        Draw camera to scene
        """
        draw_cam(self.camera_pos, self.rot_mat(), radius=0.2)

    @abstractmethod
    def update(self, window: pyglet.window.Window) -> None:
        """
        Overwrite in child to change cam during run time
        :param window: current window object
        """

    @property
    def camera_pos(self) -> np.array:
        """
        :return: camera position
        """
        return self._camera_pos.copy()

    @camera_pos.setter
    def camera_pos(self, value: Union[np.array, Tuple[float, float, float]]):
        """
        :param value: numpy array or 3 entry large tuple/list
        """
        self._camera_pos = np.array(value)

    @property
    def camera_view(self) -> np.array:
        """
        :return: camera view direction (normalized)
        """
        return self._camera_view.copy()

    @camera_view.setter
    def camera_view(self, value: Union[np.array, Tuple[float, float, float]]):
        """
        :param value: numpy array or 3 entry large tuple/list
        """
        arr = np.array(value)
        self._camera_view = arr / np.linalg.norm(arr)  # normalize vector
        self._camera_up = self._update_camera_up()

    @property
    def camera_up(self) -> np.array:
        """
        :return: camera up vector (normalized)
        """
        return self._camera_up.copy()

    @camera_up.setter
    def camera_up(self, value: Union[np.array, Tuple[float, float, float]]):
        """
        :param value: numpy array or 3 entry large tuple/list
        """
        arr = np.array(value)
        self._camera_up = arr / np.linalg.norm(arr)  # normalize vector
        self._camera_up = arr

    def rot_mat(self) -> np.ndarray:
        """
        return the rotation matrix from world to camera coordinates
        :return: rotation matrix as numpy array (3, 3)
        """
        x_c = np.cross(self.camera_view, self.camera_up)
        y_c = self.camera_up
        z_c = self.camera_view
        R = np.concatenate((x_c, y_c, z_c)).reshape(3, 3).T
        return R

    def transf(self):
        """
        Compute and return the transformation matrix from world to camera coordinates
        :return:
        """
        return construct_T(self.rot_mat(), self.camera_pos)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : Vertex array object that handles several buffer objects.
@File      : vao.py
@Time      : 31.08.21 23:50
@Author    : flowmeadow
"""

import ctypes as ct
from typing import Optional

import numpy as np
from pyglet.gl import *


class VAO:
    """
    Vertex array object that handles several buffer objects.
    """

    # TODO: not working with a combination of GL_QUADS and GL_TRIANGLES
    def __init__(self, indices: np.array, object_id: int = GL_TRIANGLES):
        """
        Initialize VAO
        (l=3: GL_TRIANGLES, l=4: GL_QUADS)
        :param object_id: can be GL_TRIANGLES or GL_QUADS
        :param indices: indices array of the model [m, l]
        """
        self.object_id = object_id  # GL_TRIANGLES or GL_QUADS
        self.vertices_per_face = indices.shape[1]

        data_type = np.uint32  # index data type
        self.index_size = data_type().itemsize  # bytes per index (default: 32 bit)

        # transform array and get data length
        indices = np.array(indices, dtype=data_type).flatten()  # flatten 2D array
        self.index_count = indices.shape[0]  # number of indices

        # generate VAO
        self.vao = GLuint()
        glGenVertexArrays(1, self.vao)
        glBindVertexArray(self.vao)

        # generate VBOs
        self.ibo = GLuint()
        self.vbo = GLuint()
        self.cbo = GLuint()
        self.nbo = GLuint()

        glGenBuffers(1, self.ibo)  # indices
        glGenBuffers(1, self.vbo)  # vertex
        glGenBuffers(1, self.cbo)  # color
        glGenBuffers(1, self.nbo)  # normals

        # bind index data
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.ibo)
        byte_length = self.index_count * self.index_size
        c_data = (ct.c_uint * self.index_count)(*indices)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, byte_length, c_data, GL_STATIC_DRAW)
        glBindVertexArray(0)

    def set_vbo(self, attr_name: str, data: np.array) -> None:
        """
        Assign vertex attributes to VAO
        (k=2: textures, k=3: vertices, k=4: colors)
        :param attr_name: name of the vertex attribute ('position', 'color', 'normal')
        :param data: array containing the vertex attribute data [n, k]
        """
        glBindVertexArray(self.vao)
        # assign per vertex position information
        if attr_name == "position":
            glEnableClientState(GL_VERTEX_ARRAY)
            glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
            self.assign_vbo_data(data)
            glVertexPointer(3, GL_FLOAT, 0, None)
        # assign per vertex color information
        elif attr_name == "color":
            glEnableClientState(GL_COLOR_ARRAY)
            glBindBuffer(GL_ARRAY_BUFFER, self.cbo)
            self.assign_vbo_data(data)
            glColorPointer(3, GL_FLOAT, 0, None)
        # assign per vertex normal information
        elif attr_name == "normal":
            glEnableClientState(GL_NORMAL_ARRAY)
            glBindBuffer(GL_ARRAY_BUFFER, self.nbo)
            self.assign_vbo_data(data)
            glNormalPointer(GL_FLOAT, 0, None)
        # TODO: Texture Coords are not defined yet. Do I miss something else?
        else:
            raise NotImplementedError("Unknown ID")
        glBindVertexArray(0)

    def draw(self, num_indices: Optional[int] = None, offset: int = 0) -> None:
        """
        Draw the VAO
        :param num_indices: number of indices that are drawn
        :param offset: defines the first index to draw
        """
        if num_indices is None:  # if nothing given, draw everything
            num_indices = self.index_count
        offset = int(offset * self.vertices_per_face)
        num_indices = int(num_indices * self.vertices_per_face)

        # bind VAO and draw elements
        glBindVertexArray(self.vao)
        glDrawElements(self.object_id, num_indices, GL_UNSIGNED_INT, ct.c_void_p(offset * self.index_size))
        glBindVertexArray(0)

    @staticmethod
    def assign_vbo_data(data: np.array) -> None:
        """
        Assign vertex attributes to VAO
        (k=2: textures, k=3: vertices, k=4: colors)
        :param data: array containing the vertex attribute data [n, k]
        """
        data_type = np.float32  # default
        # transform array and get data length
        data = np.array(data, dtype=data_type).flatten()
        data_count = data.shape[0]
        # define number of bytes and transform data
        byte_length = data_count * data_type().itemsize
        c_data = (ct.c_float * data_count)(*data)
        # assign data
        glBufferData(GL_ARRAY_BUFFER, byte_length, c_data, GL_STATIC_DRAW)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : Base window class that handles setup, catches frame rates and user inputs
@File      : base.py
@Project   : pygletPlayground
@Time      : 02.10.21 12:26
@Author    : flowmeadow
"""
from abc import abstractmethod
from typing import Tuple

import pyglet
from glpg_flowmeadow.input.mouse import Mouse
from pyglet.window import key


class Base(pyglet.window.Window):
    """
    Base window class that handles setup, catches frame rates and user inputs
    """

    # default settings for pyglet.window.Window
    _parent_kwargs = dict(
        width=None,
        height=None,
        caption=None,
        resizable=False,
        style=None,
        fullscreen=False,
        visible=True,
        vsync=True,
        file_drops=False,
        display=None,
        screen=None,
        config=None,
        context=None,
        mode=None,
    )

    # default settings for custom window
    _child_kwargs = dict(
        max_fps=120,
    )

    def __init__(self, **kwargs):
        """
        :param kwargs: keyword arguments (see class header for available keywords)
        """
        # update kwargs
        for key, val in kwargs.items():
            if key in self._parent_kwargs.keys():
                self._parent_kwargs[key] = val
            elif key in self._child_kwargs.keys():
                self._child_kwargs[key] = val
            else:
                raise KeyError(f"Invalid keyword argument {key}")

        # initialize parent class object (pyglet.window.Window)
        super().__init__(**self._parent_kwargs)

        # set attributes
        self._max_fps = self._child_kwargs["max_fps"]
        self.current_fps = 0.0
        self.display_center = (self.width // 2, self.height // 2)

        # initialize inputs
        self.mouse = Mouse()
        self.keys = set()

        # initialize
        self.init_gl()

        # set interval for drawing function
        pyglet.clock.schedule_interval(self.my_draw, 1 / self._max_fps)

    def my_draw(self, dt: float):
        """
        Update FPS, clear screen and draw next frame
        :param dt: elapsed time since last frame
        """
        self.current_fps = pyglet.clock.get_fps()
        self.clear()
        self.draw_frame()

    def on_draw(self, *test):
        """
        Override from parent as it is also called when a user input was done. Instead, on_draw_const() is used.
        """
        pass

    @staticmethod
    def run():
        """
        Run application
        """
        pyglet.app.run()

    @abstractmethod
    def init_gl(self) -> None:
        """
        To be overwritten in child class. Initialize OpenGL settings.
        """

    @abstractmethod
    def draw_frame(self) -> None:
        """
        To be overwritten in child class. Draw objects.
        """

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int):
        """
        Update mouse object
        :param x: x-position
        :param y: y-position
        :param dx: change in x-direction
        :param dy: change in y-direction
        """
        self.mouse.update(x, y, dx, dy)

    def on_key_press(self, symbol: int, modifiers: int):
        """
        Update pressed keyboard inputs
        :param symbol: keyboard symbol (e.g. a, b, c, 1, 2, 3, ...)
        :param modifiers: keyboard modifiers (e.g. ALT, SHIFT, CTRL, ...)
        """
        if symbol in [key.Q, key.ESCAPE]:
            self.close()
        self.keys.add(symbol)

    def on_key_release(self, symbol: int, modifiers: int):
        """
        Update released keyboard inputs
        :param symbol: keyboard symbol (e.g. a, b, c, 1, 2, 3, ...)
        :param modifiers: keyboard modifiers (e.g. ALT, SHIFT, CTRL, ...)
        """
        self.keys.discard(symbol)

    @property
    def size(self) -> Tuple[int, int]:
        """
        :return: window size (width, height)
        """
        return self.width, self.height



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : Camera class that allows flying freely through the scene
@File      : fly_motion.py
@Time      : 09.09.21 22:55
@Author    : flowmeadow
"""
import numpy as np
import pyglet.window
from pyglet.window import key as gl_key
from glpg_flowmeadow.transformations.methods import rotate_vec

from glpg_flowmeadow.camera.camera import Camera


class FlyMotion(Camera):
    """
    Camera class that allows flying freely through the scene. Camera can be
    controlled using the mouse and the 'WASD' keys.
    """

    def __init__(
        self,
        window,
        **kwargs,
    ):
        """
        :param display_center: the display center coordinate
        :param camera_pos: position of the camera in world coordinates [3,]
        :param camera_view: view direction of the camera in world coordinates [3,]
        :param camera_up: vector pointing up from the camera world coordinates [3,]
        """
        self._mouse = window.mouse
        self._keys = window.keys
        window.set_exclusive_mouse(True)
        super().__init__(**kwargs)

    def update(self, window: pyglet.window.Window):
        """
        Updates the camera attributes based on user inputs
        :param window: current window object
        :return:
        """
        self.update_mouse_movement()
        self.update_key_movement()

    def update_key_movement(self):
        """
        Updates camera position based on keyboard inputs
        """
        keys = self._keys
        for key in keys:
            if key in [gl_key.UP, gl_key.W]:
                self.camera_pos += 0.01 * self.camera_view  # go forward
            if key in [gl_key.DOWN, gl_key.S]:
                self.camera_pos -= 0.01 * self.camera_view  # go backward
            if key in [gl_key.RIGHT, gl_key.D]:
                side_vec = np.cross(self.camera_view, self.camera_up)
                side_vec /= np.linalg.norm(side_vec)
                self.camera_pos += 0.01 * side_vec  # go right
            if key in [gl_key.LEFT, gl_key.A]:
                side_vec = np.cross(self.camera_view, self.camera_up)
                side_vec /= np.linalg.norm(side_vec)
                self.camera_pos -= 0.01 * side_vec  # go left

    def update_mouse_movement(self):
        """
        update camera rotation based on mouse position
        """
        mouse = self._mouse

        # look up/down
        if mouse.dy != 0:
            left_vec = -np.cross(self.camera_view, self.camera_up)
            self.camera_view = rotate_vec(
                vec=self.camera_view,
                axis=left_vec,
                angle=-mouse.dy * 0.05,
            )
        # look left/right
        if mouse.dx != 0:
            self.camera_view = rotate_vec(
                vec=self.camera_view,
                axis=self.camera_up,
                angle=-mouse.dx * 0.05,
            )

        # reset mouse
        mouse.reset()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : Light source class that manages an OpenGL light source
@File      : light.py
@Time      : 10.09.21 12:52
@Author    : flowmeadow
"""
import ctypes
from typing import Union

import numpy as np
from pyglet.gl import *
from glpg_flowmeadow.rendering.methods import draw_light

# define OpenGl update function and argument for each light parameter
# TODO: add GL_SPOT_EXPONENT
update_dict = dict(
    position={"function": glLightfv, "argument": GL_POSITION},
    ambient={"function": glLightfv, "argument": GL_AMBIENT},
    diffuse={"function": glLightfv, "argument": GL_DIFFUSE},
    specular={"function": glLightfv, "argument": GL_SPECULAR},
    spot_direction={"function": glLightfv, "argument": GL_SPOT_DIRECTION},
    a_const={"function": glLightf, "argument": GL_CONSTANT_ATTENUATION},
    a_lin={"function": glLightf, "argument": GL_LINEAR_ATTENUATION},
    a_quad={"function": glLightf, "argument": GL_QUADRATIC_ATTENUATION},
    spot_cutoff={"function": glLightf, "argument": GL_SPOT_CUTOFF},
)


class Light:
    """
    Light source class that manages an OpenGL light source
    """

    # default parameter
    position = (0.0, 0.0, 0.0)
    ambient = (1.0, 1.0, 1.0, 1.0)
    diffuse = (1.0, 1.0, 1.0, 0.0)
    specular = (1.0, 1.0, 1.0, 0.0)
    spot_direction = (-1.0, 0.0, 0.0)
    a_const = 1.0
    a_lin = 0.0
    a_quad = 0.0
    spot_cutoff = 180.0

    def __init__(self, id: int, **kwargs):
        """
        :param id: OpenGL light ID (0 - 7)
        :param kwargs: parameters to update
        """
        self._id = id
        for attr, val in kwargs.items():
            self.update(attr, val)

    def update(self, key: str, data: Union[list, tuple, np.ndarray]):
        """
        Updates OpenGL light source parameter
        :param key: parameter to change
        :param data: value
        """
        fun = update_dict[key]["function"]
        arg = update_dict[key]["argument"]

        # convert to ctypes
        if isinstance(data, (list, tuple, np.ndarray)):
            data = (ctypes.c_float * len(data))(*data)
        else:
            raise ValueError(f"Value {data} for parameter {key} of light source {self._id} has wrong type")

        # update parameter
        glPushMatrix()
        glLoadIdentity()
        fun(self._id, arg, data)
        self.__setattr__(key, data)
        glPopMatrix()

    def draw(self):
        """
        Draws a representation of the light source into the scene
        """
        draw_light(self.position)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : Simple mouse class containing position and relative position change
@File      : mouse.py
@Project   : pygletPlayground
@Time      : 13.10.21 19:45
@Author    : flowmeadow
"""
import attr


@attr.s
class Mouse:
    """
    Simple mouse data structure containing position and relative position change
    """
    x: int = attr.ib(default=0)
    y: int = attr.ib(default=0)
    dx: int = attr.ib(default=0)
    dy: int = attr.ib(default=0)

    def update(self, x: int, y: int, dx: int, dy: int):
        """
        Update attributes
        :param x: x-position
        :param y: y-position
        :param dx: change in x-direction
        :param dy: change in y-direction
        """
        self.x = x
        self.y = y
        self.dx += dx
        self.dy += dy

    def reset(self):
        """
        reset relative change
        """
        self.dx = 0
        self.dy = 0



import os

from setuptools import find_packages, setup

os.system("pip freeze > requirements.txt")
REQUIREMENTS = [i.strip() for i in open("requirements.txt").readlines()]
print(REQUIREMENTS)
setup(
    name="glpg_flowmeadow",
    version="0.1.4",
    license="MIT",
    packages=find_packages(include=["glpg_flowmeadow", "glpg_flowmeadow.*"]),
    install_requires=[
        "attrs==21.4.0",
        "cycler==0.11.0",
        "fonttools==4.33.3",
        "kiwisolver==1.4.3",
        "numpy==1.22.3",
        "overrides==6.1.0",
        "packaging==21.3",
        "pyglet==1.5.23",
        "pyparsing==3.0.9",
        "python-dateutil==2.8.2",
        "six==1.16.0",
        "typing-utils==0.1.0",
    ],
    include_package_data=True,
    description="pyglet OpenGL playground",
    author="Florian Wiese",
    author_email="florian-wiese93@outlook.de",
    url="https://github.com/flowmeadow/pygletPlayground",
    download_url="https://github.com/flowmeadow/pygletPlayground.git",
    keywords=["pyglet", "OpenGL"],
)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@Introduce : General OpenGL application class, that follows a certain drawing pipeline
@File      : gl_screen.py
@Project   : pygletPlayground
@Time      : 02.10.21 13:00
@Author    : flowmeadow
"""
from abc import ABCMeta, abstractmethod

from pyglet.gl import *

from glpg_flowmeadow.camera.camera import Camera
from glpg_flowmeadow.display.base import Base


class GLScreen(Base):
    """
    General OpenGL application class, that follows a certain drawing pipeline
    """
    __metaclass__ = ABCMeta

    def __init__(self, **kwargs):
        """
        :param kwargs: forwarded keyword arguments
        """
        # TODO: How to deal with debug mode
        #       do NOT allow fullscreen only in debug mode
        #       Enable mouse in debug mode (e.g. camera.fly_motion)
        # gettrace = getattr(sys, "gettrace", None)
        # if gettrace():
        #     kwargs["fullscreen"] = False

        super().__init__(**kwargs)

        # initialize camera
        self.cam = Camera()

    def init_gl(self):
        """
        Initialize OpenGL settings
        """
        glDepthFunc(GL_LESS)
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glEnable(GL_MULTISAMPLE)

    def draw_frame(self):
        """
        Run drawing pipeline
        """
        # initialize perspective
        glLoadIdentity()
        gluPerspective(45, (self.width / self.height), 0.1, 50.0)

        # configure OpenGL settings
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        # prepare next frame
        # glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        # glClearColor(0.0, 0.0, 0.0, 0.0)
        # glClearDepth(1.0)
        # glLoadIdentity()

        # handle the gathered events if necessary (empty by default)
        self.cam.update(self)
        self.handle_events()
        glFlush()

        # initialize camera
        gluLookAt(
            *self.cam.camera_pos,
            *(self.cam.camera_pos + self.cam.camera_view),
            *self.cam.camera_up,
        )
        # draw world objects (empty by default)
        self.draw_world()
        glFlush()

        # draw on screen
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        glDisable(GL_DEPTH_TEST)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(0, self.width, 0.0, self.height, -1.0, 1.0)
        self.draw_screen()
        glEnable(GL_DEPTH_TEST)

        # # TODO: Check computational cost of this. It can be optional
        # glDisable(GL_DEPTH_TEST)
        # # compute inverse model view projection matrix
        # glLoadIdentity()
        #
        # glGetFloatv(GL_PROJECTION_MATRIX, self._projection_matrix)
        # glGetFloatv(GL_MODELVIEW_MATRIX, self._view_matrix)
        # self.inverse_matrix = np.linalg.inv(to_matrix(self._view_matrix) @ to_matrix(self._projection_matrix))
        #
        # # draw screen objects (empty by default)
        # glFlush()
        # glEnable(GL_DEPTH_TEST)

    @abstractmethod
    def handle_events(self) -> None:
        """
        handle pygame events and do other stuff before drawing
        :return: None
        """
        pass

    @abstractmethod
    def draw_world(self) -> None:
        """
        draw objects in the world
        :return: None
        """
        pass

    @abstractmethod
    def draw_screen(self) -> None:
        """
        draw objects onto the screen (e.g. GUI, text, images).
        :return: None
        """
        pass
